/** @file UNICOS E-mail/SMS handler

 	This script allows to send the SMS/email notifications due to alarms or reports,
 	as configured by the user. A limit on the number of messages may be configured.
 	The content of the notification depends on the destination address:
 	if it is using the mail2sms.cern.ch gateway as the recipient, a compact form
 	is sent. The content may be generated by a user-provided function, or extracted
 	from the "SMS_MESSAGE" device property. The timestamp, value, text of the alarm
 	as well as the source (device alias or comment) is also notified.

 	The script is executed inside the CTRL manager that runs the unicos_scripts.lst
 	list of scripts.

 	The workflow of the mechanism is actually pretty complex and it based on the
 	alert actions defined in UNICOS-Specific alert classes.
 	NOTE: alert actions are going to be deprecated in WinCC OA in future!

 	It is as follows:
 	- When an alarm on a one of UNICOS DPEs (with one of special UNICOS alert classes
 	  such as _unAnalogH_Mail) is triggered,
 	  then one of the alert actions of the alarm class is executed by the event manager(!)
 	- the alert action contains the CTRL code such as
 ```
 main(string sDp, time t, int i, string sText, float fValue, bit32 b32State)
 {
   int iRes;
   string sDataPoint, sTime, sMessage;
   string sMaskText = getBit(b32State, 11) ? " (Masked)" : "";
   sTime=formatTime("%x  %H:%M:%S", t," (milli: %03d)");
   sMessage=sDp + "~" + sTime + "~" + sText +sMaskText+ "~" + fValue;
   sDataPoint=getSystemName() + "_unSendMail.message";
   iRes=dpSet(sDataPoint + ".mail_text", sMessage);
}
 ```
  	i.e. it formats a notification message and sends it with dpSet
  	to the "maibox" DP "_unSendMail.message.mail_text
  .
  - The unSendMail.ctl script listens on this "mailbox" DP,
  (with dpConnect) and receives callbacks to _unSendMail_Message_CallBack
  - The callback processes the message, which contains the information
  on the origin dpe that caused the alarm as well as the additional info
  contained in this message, and then invokes emailing facilities
  - The actual emails (and SMS) are sent with startThread()

Configuration is stored in the elements of the _unSendMail datapoint's ".config" branch

  - mail_settings : dyn_string with settings of emails separated by the ~ (tilde) character;
   each element of the dyn_string corresponds to a single category
   Example: unProcessAlarm~cern.ch~cernmx.cern.ch~CRYO@CS-CCR-Q4DS2.cern.ch~person1@cern.ch;person2@cern.ch;+41754110000@mail2sms.cern.ch~http://abinf1/lemon/info.php?host=cs-ccr-q4ds2~250
   It contains: the category, the domain and SMTP server, the sender address, the list of recipients emails as well as SMS (with the mail2sms gateway)
   separated by semicolons, the URL, and the max number of daily notifications

  - mail_disable : dyn_bool indicating whether the notification for a category is disabled; the order follows the one of categories
  set in .config.mail_settings

  - report_settings: dyn_string: tilde-separated list of datapoints or dp name patterns used for reporting;
  corresponds to the order of categories as defined in the .config.mail_settings
  Example entry: P4_42:un-CFP_UX45_QUIC-QUIC-CPC_DigitalAlarm-00183.ProcessInput.ISt~_unSystemAlarm_*.alarm

  - reportConfiguration: dyn_string defining the list of various functions used for configuration; note that
  there is no correspondence to the .config.mail_settings, but rather a separate formatting convention. Each entry
  in the list is composed of the category name, followed by a keyword (_REPORT,_STARTUP, _MAIL which corresponds
  to the consts UN_SEND_MAIL_KEY_CATEGORY_REPORT_EXTENSION, UN_SEND_MAIL_KEY_CATEGORY_STARTUP,UN_SEND_MAIL_KEY_CATEGORY_MAIL_EXTENSION),
  then a colon (:) and then the parameter.
  For the *_STARTUP the value is either TRUE or FALSE
  For the *_REPORT and *_MAIL it is the name of the configuration function to be used (see also below)
  There is also a dedicated entry that could be present which starts with GETCATEGORY
  (corresponding to the UN_SEND_MAIL_KEY_GET_CATEGORY constant) , then a colon,
  and the name of the function; if present it determines the name of a function used to generate
  the whole configuration.
  Example of entries in the list
    MyTest01_STARTUP:TRUE
    MyTest01_REPORT:MyReportFunction
    MyTest01_MAIL:MyMailFunction
    MtTest02_STARTUP:FALSE
    GETCATEGORY:MyConfigFunction

  - .config.mail_text : a dyn_string, where every entry defines the custom notification message; note that
   the format of entry is the category name, followed by ~ (tilde) and then the actual multi-line text that could
   occupy subsequent entries (for multi-line texts).
   Example:
   	- MyTest01~This is my
   	-  multi-line message
   	-  for notifications
   	- MyTest02~ Another notification

Reporting part of the notifications is implemented through the WinCC OA Scheduler module, and hence its configuration
is stored in the dedicated datapoints of type _Sc, the names of which correspond to the categories, but it is not
the rule (some have generic names such as Scheduler_0000), and the name is in the .name element; anyway, they could
be easily reviewed (with stats of their last execution) from the SysMgm->Settings->Scheduler in GEDI.

The mechanism is the following: a timed schedule is set to trigger the _unSendMail.message.periodic_report DPE
at the defined periods of time, with the value set to the name of the category.
This is then catched by the _unSendMail_Report_CallBack in this script and handled as necessary.

The functions that could be specified for the configuration, with their arguments and return values
are described below. Note that it is the responsability of their author to ensure that they are loaded
by this CTRL script (e.g. through the appropriate LoadCtrlLibs statement in the config file)

1) the "GET_CATEGORY" configuration function allows to define which categories should be processed.
	It should accept a single string parameter, to which the notificarion information of the callback
	will be passed (e.g. "P4_42:_unSystemAlarm_DS_Comm_SOFT_FE_SOFT_FE_QBS_P4.alarm:_alert_hdl.2~11/02/2020 18:20:44.541~Bad~10")
	and it should return the dyn_string list of names of categories that should be processed.
	e.g. ```dyn_string MyGetCategory(string info)```

2) the EMAIL/SMS formatting function for a specified category and notification info;
	it should have the following signature:
	```dyn_dyn_string MyEmailSMS(string category, string dpe, string info);```
	.
	The info parameter is the one passed in the callback, e.g.
	```P4_42:_unSystemAlarm_DS_Comm_SOFT_FE_SOFT_FE_QBS_P4.alarm:_alert_hdl.2~11/02/2020 18:20:44.541~Bad~10```
	The dpe parameter is the one that originally triggered the SMS/Email callback, hence typically not useful.
	The function should return a 2x2 dyn_dyn_string array with titles for MAIL and SMS in the first row,
		// and the lines of email body in the second row (as columns)
		// [1][1] -> SMS Title   [UN_SEND_MAIL_TITLE=1][SMS_TITLE=1]
		// [1][2] -> EMail Title [UN_SEND_MAIL_TITLE=1][MAIL_TITLE=2]
		// [2][*] -> Successive lines for the Email body [UN_SEND_MAIL_BODY=2]

3) the REPORT formatting function: it is similar to the EMAIL/SMS function above,
	yet its signature should be
	```dyn_dyn_string MyReport(string category, dyn_string dpeList)```
	whereby the dpeList param will get the list of DPEs the value of which should
	be reported. The format of the return value is as in the EMAIL/SMS function above.

 */

#uses "unSystemIntegrity/unSendMail.ctl"

//@{

// global declaration
dyn_string g_dsSender;					// Mail address for the Sender
dyn_string g_dsType;					// Type of Alarm
dyn_bool g_dbMax;						// Max number of Emails reached
dyn_string g_dsReceiver;				// Mail address(es) for the Receiver
dyn_string g_dsSmtp;        			// smtp server
dyn_string g_dsDomain;			      	// Domain name
dyn_string g_dsUser;						//User name
dyn_string g_dsPassword;				//Password name
dyn_bool g_dbUseSsl;						//Use SSL/TLS when sending email
dyn_string g_dsLink;					// Link name (lemon for ex)
dyn_string g_dsMax;						// max number of mails per day and per receiver(=mailbox)
dyn_string g_dsDpe;						// Report Dpe list
dyn_bool g_dbDisable;					// True = Disable Sending Mails
mapping g_diDayCounter; 				// number of mail sent per category and per day
mapping g_diToday; 						// day number per category

global string g_Mail_sSystemName;

const string SMS_PATTERN = "*@mail2sms.cern.ch";



const int DPE=1;
const int TIME=2;
const int ALERT_TEXT=3;
const int VALUE=4;
const int DESC = 4;

const int DELAY_RESEND=60;
const int SEND_ATTEMPTS=2;
const int DELAY_INIT=5;

const string c_DPE_Backup = "_unApplication.unBackupStatus";


struct UnSendMailServerInfo {
	string category;
	string sendFrom;
	string sendTo;
	string smtpServer;
	string domain;
	string user;
	string password;
	bool useSsl;

	public UnSendMailServerInfo(string sCategory, string sFrom, string sTo, string sServer, string sDomain, string sUser, string sPassword, bool bUseSsl) {
		category=sCategory;
		sendFrom=sFrom;
		sendTo=sTo;
		smtpServer=sServer;
		domain=sDomain;
		user=sUser;
		password=sPassword;
		useSsl=bUseSsl;
	}

	/**
		Construct the info based on the category index (which must be valid),
		and filter-in either SMSes or emails

		@par idx category index in the global variables; no error checking is done!
		@par selectSMS indicates if one wants the SMS (true) or emails (false)
	*/
	public static UnSendMailServerInfo fromCatIdx(int idx, bool selectSMS) {

		dyn_string dsReceiver=strsplit(g_dsReceiver[idx],UN_SENDMAIL_DELIMITER);
		string sTo;
		for (int i=1;i<=dynlen(dsReceiver);i++) {
			bool isSMS=patternMatch(SMS_PATTERN, dsReceiver[i]);
			if ( (isSMS && selectSMS) || ( (!isSMS) && (!selectSMS) ) ) sTo=sTo+dsReceiver[i]+";";
		}
		return UnSendMailServerInfo(g_dsType[idx],g_dsSender[idx],sTo,g_dsSmtp[idx],g_dsDomain[idx], g_dsUser[idx], g_dsPassword[idx], g_dbUseSsl[idx]);
	}
};

// mapping of error codes from the unSendMail.cat - must be numeric
private const int INFO_GENERIC            = 0;
private const int EXC_DP_DOESNT_EXIST     = 1;
private const int EXC_CANNOT_GET_SETTINGS = 2;
private const int EXC_INCOHERENT_SETTINGS = 3;
private const int EXC_CANNOT_DP_CONNECT   = 4;
private const int EXC_WRONG_USERFUNCTION  = 5;
private const int EXC_WRONG_MAIL_PARAMS   = 6;
private const int EXC_SEND_FAIL           = 7;
private const int EXC_EMPTY_CATEG_FILTER  = 8;
private const int EXC_INCOHERENT_DATA     = 9;

/**
 * Purpose: This is the main script.
 */
int main()
{

	string sDp_1, sDp_2;
	dyn_bool dbDisable;
	dyn_string dsSettings, exceptionInfo, dpToCheck, dsVersion, dsReport;
	int iRes, len_1, len_2;
	dyn_dyn_string ddsComponents;
	int i, len;
	dyn_errClass err;

	g_Mail_sSystemName = getSystemName();

	sDp_1="_unSendMail.config";
	sDp_2="_unSendMail.message";

	delay(DELAY_INIT);

try {

	if ( (!dpExists(sDp_1)) || (!dpExists(sDp_2)) ) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_DP_DOESNT_EXIST,sDp_1, sDp_2));

	// check coherency of settings
	iRes=dpGet(sDp_1 + ".mail_settings", dsSettings, sDp_1 + ".mail_disable", dbDisable, sDp_1 + ".report_settings", dsReport);
	err=getLastError();
	if (iRes<0 || dynlen(err)) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_CANNOT_GET_SETTINGS,sDp_1, getErrorText(err)));

	len_1=dynlen(dsSettings);
	len_2=dynlen(dbDisable);

	if ( (len_1!=len_2) ) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_INCOHERENT_SETTINGS,len_1,len_2));

	if (dynlen(dsReport)!=len_1)	{
		//check coherency and adjust if needed
		while(dynlen(dsReport)<len_1) dynAppend(dsReport, "");
		while(dynlen(dsReport)>len_1) dynRemove(dsReport, dynlen(dsReport));
		dpSetWait(sDp_1 + ".report_settings", dsReport);
		fwException_raise(exceptionInfo, "INFO", "_unSendMail.report_settings dpe automatically updated - dynlen size was not correct!","");
		_handleExcInfo(exceptionInfo);
	}

	// Set up callbacks

	iRes = dpConnect("_unSendMail_Settings_CallBack", sDp_1 + ".mail_settings", sDp_1 + ".mail_disable",sDp_1 + ".report_settings");
	if (iRes < 0) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_CANNOT_DP_CONNECT,sDp_1,"_unSendMail_Settings_CallBack"));

	iRes = dpConnect("_unSendMail_Message_CallBack", false, sDp_2 + ".mail_text");
	if (iRes < 0) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_CANNOT_DP_CONNECT,sDp_1,"_unSendMail_Message_CallBack"));

	iRes = dpConnect("_unSendMail_Report_CallBack", false, sDp_2 + ".periodic_report");
	if (iRes < 0) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_CANNOT_DP_CONNECT,sDp_1,"_unSendMail_Report_CallBack"));


	string compVersions = unGenericDpFunctions_getComponentsVersion(makeDynString("unSystemIntegrity", "unCore"));
	throwError(makeError("unSendMail",PRIO_INFO,ERR_CONTROL,INFO_GENERIC,"started "+compVersions));

	unSendMail_SendInitNotifications();

} catch {
	_handleErrClass(getLastException());
	throwError(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,INFO_GENERIC,"stopped on exceptions; may not work correctly"));
}

}

/** Exception handling function

	Send the exception passed in the @err parameter to the unMessageText and
	also rethrows it as an error to the WinCC OA log, with proper formatting

	@par err exception that is to be handled
*/
void _handleErrClass(const dyn_errClass &err)
{
	string errTxt=getErrorText(err);
	string errPrio="";
	switch(getErrorPriority(err)) {
		case PRIO_FATAL:
		case PRIO_FATAL_NO_RESTART: errPrio="FATAL";  break;
		case PRIO_SEVERE:           errPrio="ERROR";  break;
		case PRIO_WARNING:          errPrio="WARNING";break;
		case PRIO_INFO:             errPrio="INFO";   break;
	}
	dyn_string exceptionInfo; // dummy parameter needed below
	unMessageText_send("*", "*", getSystemName()+"SendMail", "user", "*", errPrio, errTxt, exceptionInfo);
	throwError(err); // and make sure to send it to the WinCC OA log too
}

/** Sends a email with specific params, subject and text

	@par srv 			the UnSendMailServerInfo parameter properly initialized for Email sending
	@par emailSubject 	the subject of the email
	@par emailText 		the content of the email
	@par type			the "type" of notification, usually being the category; optional,
						with default ("") resolving to the category defined in the @c srv parameter
	@return true on success (ie. if a thread was spawned; still, the sending or delivery might fail downstream)
 */
bool _unSendMail_sendEmail(const UnSendMailServerInfo &srv, const string &emailSubject, const string &emailText, string type="")
{
	if (type=="") type=srv.category;
	int iThread=startThread("_unSendMail_Message", srv.sendTo, srv.sendFrom, srv.smtpServer, srv.domain, emailSubject, emailText, type, srv.user, srv.password, srv.useSsl);
	return (iThread>=0);
}

/** Sends a SMS with specific params and text

	@par srv 			the UnSendMailServerInfo parameter properly initialized for SMS sending
	@par smsText 		the content of the SMS
	@par type			the "type" of notification, usually being the category; optional,
						with default ("") resolving to the category defined in the @c srv parameter
	@return true on success (ie. if a thread was spawned; still, the sending or delivery might fail downstream)
 */
bool _unSendMail_sendSMS(const UnSendMailServerInfo &srv, const string &smsText, string type="")
{
	if (type=="") type=srv.category;
	string emptyBody="";
	int iThread=startThread("_unSendMail_Message", srv.sendTo, srv.sendFrom, srv.smtpServer, srv.domain, smsText, emptyBody, srv.category, type, srv.user, srv.password, srv.useSsl);
	return (iThread>=0);
}


/** Handle the exceptionInfo

	The function checks the exceptionInfo variable passed as the argument,
	and sends it to the unMessageText channel (if available) or prints it out to the log.
	Then it clears the exceptionInfo variable.
	Note that multiple exceptions in the exceptionInfo are handled correctly

	For convenience, the returns TRUE if the exceptionInfo was not empty.

	@par exceptionInfo : the standard exception handling variable
	@return FALSE if the exceptionInfo was empty, TRUE otherwise

*/
private bool _handleExcInfo(dyn_string &exceptionInfo)
{
	if (dynlen(exceptionInfo)==0) return false;

	// throw in the log, using nice error messages
	for (int i=0;i<(dynlen(exceptionInfo)/3);i++) {

		int prio=PRIO_INFO;
		if (exceptionInfo[3*i+1]=="WARNING") prio=PRIO_WARNING;
		else if (exceptionInfo[3*i+1]=="ERROR") prio=PRIO_SEVERE;

		throwError(makeError("unSendMail",prio,ERR_CONTROL,INFO_GENERIC,exceptionInfo[3*i+1]+":"+exceptionInfo[3*i+2]));
	}

	if (isFunctionDefined("unMessageText_sendException")) {
		unMessageText_sendException("*", "*", g_Mail_sSystemName+"SendMail", "user", "*", exceptionInfo);
	}

	dynClear(exceptionInfo);
	return true;
}

/** Callback function called when _unSendMail.message.mail_text is triggered
 *
 *	@param sDataPoint_1: the DP that triggers the callback ("_unSendMail.message.mail_text")
 *	@param sValue_1: string, input, Information; it has a format such as
 *		"P4_42:_unSystemAlarm_DS_Comm_SOFT_FE_SOFT_FE_QBS_P4.alarm:_alert_hdl.2~11/02/2020 18:20:44.541~Bad~10"
 *
 */
void _unSendMail_Message_CallBack(string sDataPoint_1, string sValue_1)
{

try {
	string sFunction;
	dyn_string exceptionInfo;

	// determine if there is the overall GET_CATEGORY configuration function defined
	unProcess_getReportCategoryFunction("", UN_SEND_MAIL_KEY_GET_CATEGORY, sFunction, exceptionInfo);

	if (sFunction == "") {
		// default handling of all categories with their configurations
		_unSendMail_Mail(sValue_1);
		return;
	}

	if (isFunctionDefined(sFunction)) {
		dyn_string dsCategoryFilter;
		// get the configuration using the function
		evalScript(dsCategoryFilter,
					"dyn_string main(string sValue) "+
					"{" +
					"return " + sFunction + "(sValue); "+
					"}",
					makeDynString(), sValue_1);
		// the function returns the list of categories that should be treated
		// process them...
		_unSendMail_Mail(sValue_1, dsCategoryFilter);

	} else {
		// no function defined - report an error
		string sDpe = sValue_1;
		// to find the DPE, parse the data
		dyn_string dsTempSplit=strsplit(sValue_1, UN_SENDMAIL_SEPARATOR);
		if (dynlen(dsTempSplit)>=4) {
			sDpe=dsTempSplit[DPE];
			sDpe=substr(sDpe,0,strpos(sDpe,":_alert_hdl"));
		}
		throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_WRONG_USERFUNCTION,sFunction, sDpe));
	}

} catch {
	_handleErrClass(getLastException());
	throwError(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,INFO_GENERIC,"stopped on exception."));
}

}

//------------------------------------------------------------------------------------------------------------------------

/** send a Mail/SMS
 *
 * @param sValue_1  input, the data payload of the callback, ie. the tilde-separated notification data
 * @param categoriesFilter additional filter allowing to select the categories to be processed;
 * 			used namely when the custom category-configuration function is employed.
 * 			The parameter should contain a list of category names; leaving the default,
 * 			i.e. a single element in the list with "*" means all categories should be processed.
 */
void _unSendMail_Mail(string sValue_1, dyn_string categoriesFilter=makeDynString("*"))
{
	dyn_string dsReceiver, exceptionInfo, dsTempSplit, dsDpe, dsDp_toCheck, dsText, exceptionInfoTemp;
	int iRes, iIndex, i, j, k, len, len_check, len_dpe, len_receiver, iProcess, iMax, iDay;
	string sSender, sDomain, sMessage, sTime, sAlert_text, sDpe, sDescription, sValue, sReceivers, sTitleMail, sTitleSMS, sBody, sMain, sText,sDesc, sLink;
	string sReceiverSMS, sReceiverMail, sAlias, sCustomMessage;
	string sFunction;
	dyn_dyn_string ddsRetFunction;
	string sAlertMessage,sDpName;
	dyn_string dsConfig;

	if (sValue_1=="") return; // invalid trigger; just silently skip it

	if (dynlen(categoriesFilter)==0) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_EMPTY_CATEG_FILTER));

	// process the categoriesFilter for possible regexp patterns
	dyn_string selectedCategories=g_dsType; // the default...
	if (categoriesFilter[1]!="*" || dynlen(categoriesFilter)>1) {
		dynClear(selectedCategories);
		for (i=1;i<=dynlen(categoriesFilter);i++) {
			dyn_string pmatch=dynPatternMatch(categoriesFilter[i],g_dsType);
			dynAppend(selectedCategories,pmatch);
		}
		dynUnique(selectedCategories);
	}

	if (dynlen(selectedCategories)<1) return; // nothing to do really - e.g. no categories at all


	// parse the data
	dsTempSplit=strsplit(sValue_1, UN_SENDMAIL_SEPARATOR);
	if (dynlen(dsTempSplit)!=4) throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_INCOHERENT_DATA,sValue_1));

	sDpe=substr(dsTempSplit[DPE],0,strpos(dsTempSplit[DPE],":_alert_hdl"));
	sTime=dsTempSplit[TIME];
	sAlert_text=dsTempSplit[ALERT_TEXT];
	sValue=dsTempSplit[VALUE];
	sText=_unSendMail_convertValueToText(sDpe, sValue);

	iDay=day(getCurrentTime());

	//Description or Alias?
	iProcess=strpos(sDpe,"ProcessInput.");
	if (iProcess>=0) {
		sDescription=dpGetAlias(substr(sDpe,0,iProcess));
		if (sDescription=="") sDescription=sDpe;
	} else {
		sDescription=unGenericDpFunctions_getDescription(sDpe);
	}

	//Parameters
	len=dynlen(g_dsType);
	for (i=1;i<=len;i++) {	// for each Category

		string category=g_dsType[i];
		// the key used in mappings
		string cat=strtolower(category);

		// check if we should process this category based on filtering
		if (!dynContains(selectedCategories,category))  return;

		// msg counter: check if it is a new day
		if (g_diToday[cat] != iDay) {   // new day => reset the counter
			g_diToday     [cat] = iDay;
			g_diDayCounter[cat] = 0;
		} else {
			_unSendMail_checkMapping(category);	// check/init the mapping definition
		}

		if (dynlen(g_dbDisable)<i  || dynlen(g_dsSender)<i ||
			dynlen(g_dsReceiver)<i || dynlen(g_dsDomain)<i) {
			sMessage="_unSendMail_Mail(): " + getCatStr("unSendMail", "PARAMETERS_MISSING") + dsDpe + " - " + i + " - " + dynlen(g_dbDisable) + " - " + dynlen(g_dsSender) + " - " + dynlen(g_dsReceiver) + " - " + dynlen(g_dsDomain);
			fwException_raise(exceptionInfo, "ERROR", sMessage,"");
			continue;
		}

		// check conditions
		if ( ((int)g_diDayCounter[cat]) > ((int)g_dsMax[i]) ) continue; // exceeded the number of notifications per day;
		if (g_dbDisable[i]) continue; // disabled

		UnSendMailServerInfo emailParams=UnSendMailServerInfo::fromCatIdx(i,false);
		UnSendMailServerInfo smsParams=UnSendMailServerInfo::fromCatIdx(i,true);

		dsDpe=strsplit(g_dsDpe[i],UN_SENDMAIL_SEPARATOR);
		len_dpe=dynlen(dsDpe);

		sLink =g_dsLink[i];

		// process wildcards and aliases in monitored DPs
		dynClear(dsDp_toCheck);
		for (int n=1;n<=len_dpe;n++) {
			dyn_string dsDPtoAppend;
			if (strpos(dsDpe[n],"@") == 0) {
				dyn_string dsAliases;
				dpGetAllAliases(dsDPtoAppend,dsAliases,substr(dsDpe[n],1));
			} else {
				dsDPtoAppend=dpNames(dsDpe[n]);
			}
			if (dynlen(dsDPtoAppend)>0) dynAppend(dsDp_toCheck,dsDPtoAppend);
		}
		dynUnique(dsDp_toCheck);
		len_check=dynlen(dsDp_toCheck);

		// check if amongst the monitored DPEs in the current category,
		// there is the one that triggered us; we know that the list
		// is unique - hence contains() is enough!
		// if not found - just pass to the next one;
		if (!dsDp_toCheck.contains(sDpe)) continue;

		// process the case when we've just reached the daily limit
		if ( ((int)g_diDayCounter[cat]) == ((int)g_dsMax[i]) )	{

			sMessage="_unSendMail_Mail(): " + getCatStr("unSendMail", "TOO_MUCH_MAIL") + " (" + iMax + ") - Category: " + category;
			fwException_raise(exceptionInfo, "ERROR", sMessage,"");

			if (emailParams.sendTo!="") {
				dynClear(dsText);
				dynAppend(dsText, "Please check the application.\n");
				dynAppend(dsText, "The Max number of Emails (" + iMax + ") has been reached for the day. This category is disabled until midnight.\n");
				_unSendMail_getText("WARNING", category, category, sLink, dsText, sTitleMail, sBody);
				_unSendMail_sendEmail(emailParams, sTitleMail, sBody);
			}

			if (smsParams.sendTo!="") {
				sTitleSMS=_unSendMail_getTitle("WARNING", "Exceeded SMS for today:"+category);
				_unSendMail_sendSMS(smsParams, sTitleSMS);
			}

			continue;
		}



		// retrieve the custom function name
		dynClear(ddsRetFunction);
		sFunction = "";
		dynClear(exceptionInfoTemp);
		unProcess_getReportCategoryFunction("", category+UN_SEND_MAIL_KEY_CATEGORY_MAIL_EXTENSION, sFunction, exceptionInfoTemp);
		dynAppend(exceptionInfo, exceptionInfoTemp);

		// process the custom-function for notification content
		if (sFunction != "") {
			if (isFunctionDefined(sFunction)) {
				evalScript(	ddsRetFunction,
							"dyn_dyn_string main(string sCategory, string sDpe, string sValue) "+
							"{" +
							"return " + sFunction + "(sCategory, sDpe, sValue); "+
							"}",
							makeDynString(), category, sDpe, sValue);
				// pad with the body part for the case where the function returned only the titles
				while (dynlen(ddsRetFunction) <= UN_SEND_MAIL_BODY) dssRetFunction.append(makeDynString("", ""));

				sTitleMail = ddsRetFunction[UN_SEND_MAIL_TITLE][MAIL_TITLE];
				sTitleSMS  = ddsRetFunction[UN_SEND_MAIL_TITLE][SMS_TITLE];
				sBody = _unSendMail_convertBody(ddsRetFunction[UN_SEND_MAIL_BODY]);
			} else {
				fwException_raise(exceptionInfo, "ERROR", "_unSendMail_Mail(): user function not defined: " + sFunction,"");
			}
		} else {
			// standard case: format the notifications ourselves
			// UNTOOL-28: get the alarm messages from device configuration
			sAlertMessage = "";
			_unGenericDpFunctions_getKeyConfiguration(dpAliasToName(sDescription),"statusInformation.configuration.deviceConfiguration","SMS_MESSAGE",dsConfig,exceptionInfo);
			if (dynlen(dsConfig) >=1) sAlertMessage = dsConfig[1];

			// format the Email into sTitleMail and sBody
			dynClear(dsText);
			dynAppend(dsText, "Value=" + sText + " " +sAlertMessage);
			_unSendMail_getText(sAlert_text, sDescription, category, sLink, dsText, sTitleMail, sBody, sTime);

			// format the SMS into sTitleSMS
			sTitleSMS=_unSendMail_getTitle(sAlert_text, sDescription,sTime);
			// we do not pass the alias to save space
			sTitleSMS =  sTitleSMS + " Value=" +sText +" " +sAlertMessage;
		}

		bool ok1,ok2;
		if (emailParams.sendTo!="") ok1 = _unSendMail_sendEmail(emailParams, sTitleMail, sBody);
		if (smsParams.sendTo!="")   ok2 = _unSendMail_sendSMS(smsParams, sTitleSMS);

		// update the counter of daily messages
		if (ok1||ok2)  g_diDayCounter[cat]++;

	} // end of the for-loop over categories

	if (_handleExcInfo(exceptionInfo)) return;
}


//------------------------------------------------------------------------------------------------------------------------

/**
 * Purpose: Callback function to save settings
 *
 *	@param sDataPoint_1: string, input, data point name
 *	@param dsSettings: dyn_string, input, Settings: Type / From / To
 *	@param sDataPoint_2: string, input, data point name
 *	@param dbDisable: dyn_string, input, Disable Sending?
 *
 */
void _unSendMail_Settings_CallBack(string sDataPoint_1, dyn_string dsSettings, string sDataPoint_2, dyn_string dbDisable, string sDataPoint_3, dyn_string dsReport)
{
		synchronized (g_dsType)
	{
		unSendMail_parseConfigMailSettings(dsSettings, g_dsType, g_dsSender, g_dsReceiver, g_dsDomain, g_dsSmtp, g_dsLink, g_dsMax, g_dsUser, g_dsPassword, g_dbUseSsl);

    //UNCORE-240 update  g_diToday with proper entries
    for(int i=1; i<=dynlen(g_dsType); i++){
      g_diToday[strtolower(g_dsType[i])]=0;
    }

		g_dbDisable=dbDisable;

		dynClear(g_dsDpe);
		g_dsDpe=dsReport;
	}
}

/** Purpose: Callback function to send a report
 *
 *	@param sDataPoint_1: string, input, data point name
 *	@param sType: string, input, scheduler Type
 *
 */
void _unSendMail_Report_CallBack(string sDataPoint_1, string sType)
{
	try {
		string sFunction;
		dyn_string exceptionInfo;
		unProcess_getReportCategoryFunction("", sType+UN_SEND_MAIL_KEY_CATEGORY_REPORT_EXTENSION, sFunction, exceptionInfo);
		if (sFunction == "") {
			_unSendMail_Report(sType);
		} else {
			_unSendMail_ReportUserFunction(sType, sFunction);
		}
	} catch {
		_handleErrClass(getLastException());
		throwError(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,INFO_GENERIC,"stopped on exception."));
	}
}

//------------------------------------------------------------------------------------------------------------------------

/** send a report
 *
 *
 * @param sType  input, the mail/SMS category
 */
void _unSendMail_Report(string sType)
{
	dyn_string dsDpe, dsDp_toCheck, dsReceiver, exceptionInfo, dsText;
	string sText, sTitle, sMessage, sDp, sDpe, sAlarm, sAlias, sInformation, sValue, sBlank_1, sAck, s1, s2, s3, sSender, sDomain, sSmtp, sLink;
	string sBad, sBackup, sBody, sReceiverSMS, sReceiverMail;
	int iIndex, iThread, iRes, len_check, len, i, j, len1, len10, len20, len30, iProcess, iBad = 0;


	sBlank_1="                                    ";   //use to format text message (len=9)
	len1=strlen(sBlank_1);

	iIndex=dynContains(g_dsType, sType);
	if (iIndex <=0) {
		fwException_raise(exceptionInfo, "ERROR", "_unSendMail_Report(): " + getCatStr("unSendMail", "UNKNOWNCATEGORIE") + sType,"");
		_handleExcInfo(exceptionInfo);
		return;
	}

	sLink = g_dsLink[iIndex];

	// check if backup is ok
	if (dpExists(c_DPE_Backup)) {
		string sLastBackup, sTimeLastBackup;
		time tTimeLastBackup;

		dpGet(c_DPE_Backup, sLastBackup, c_DPE_Backup+":_online.._stime", tTimeLastBackup);
		sTimeLastBackup = (string)tTimeLastBackup;

		if ((getCurrentTime()-tTimeLastBackup)>= 86400) {
			len10=strlen("Bad");
			len20=strlen("Not executed");
			s2=substr(sBlank_1, 0, len1 - len20);
			sBackup= "         - " + " Bad\t   " + "Not executed" + s2 + "Online Backup";
		} else {
			if (sLastBackup!="") {
				len10=strlen(sLastBackup);
				len20=strlen(sTimeLastBackup);
				s2=substr(sBlank_1, 0, len1 - len20);
				sBackup= "         - " + " " + sLastBackup + "\t   " + sTimeLastBackup + s2 + "Online Backup";
			}
		}
	}

	dynAppend(dsText, "\tAlert\t  Value\t\t\t\t Alias\t\t\t\tName\n\n");
	if (sBackup!="") dynAppend(dsText, sBackup + "\n");


	if (dynlen(g_dbDisable)<iIndex || dynlen(g_dsSender)<iIndex || dynlen(g_dsReceiver)<iIndex || dynlen(g_dsDomain)<iIndex) {
		fwException_raise(exceptionInfo, "ERROR", "_unSendMail_Report(): " + getCatStr("unSendMail", "PARAMETERS_MISSING"),"");
		_handleExcInfo(exceptionInfo);
		return;
	}

	if (g_dbDisable[iIndex]) return;

	dsDpe=strsplit(g_dsDpe[iIndex],"~");
	len=dynlen(dsDpe);
	dynClear(dsDp_toCheck);
	const string ALIAS_PLACEHOLDER="AliasPattern:";
	for (int n =1;n<=len;n++) {
		dyn_string dsDPtoAppend;
		if (strpos(dsDpe[n],ALIAS_PLACEHOLDER) == 0) {
			//DebugN("AliasPattern,",substr(dsDpe[n],13));
			dyn_string dsAliases;
			dpGetAllAliases(dsDPtoAppend,dsAliases,substr(dsDpe[n],dynlen(ALIAS_PLACEHOLDER)));
		} else {
			dsDPtoAppend=dpNames(dsDpe[n]);
		}

		if (dynlen(dsDPtoAppend)>0) dynAppend(dsDp_toCheck,dsDPtoAppend);
	}

	dynUnique(dsDp_toCheck);
	len_check=dynlen(dsDp_toCheck);

	for (j=1;j<=len_check;j++) {	//for each dpe
		sValue="";
		sAlarm="";
		sInformation="";

		//Get Value and Alert Text
		_unSendMail_Status_Value(dsDp_toCheck[j], sValue, sAlarm, sInformation);
		sText=_unSendMail_convertValueToText(dsDp_toCheck[j], sValue);
		//Description or Alias?
		iProcess=strpos(dsDp_toCheck[j],"ProcessInput.");
		if (iProcess>=0) {
			sDpe=dpGetAlias(substr(dsDp_toCheck[j],0,iProcess));
			if (sDpe=="") sDpe=dsDp_toCheck[j];
		} else {
			sDpe=unGenericDpFunctions_getDescription(dsDp_toCheck[j]);
		}

		sAlias = dpGetAlias(dsDp_toCheck[j]);
		len10=strlen(sAlarm);
		len20=strlen(sText);
		len30=strlen(sAlias);
		s1=substr(sBlank_1, 0, len1 - len10);
		s2=substr(sBlank_1, 0, len1 - len20);
		s3=substr(sBlank_1, 0, len1 - len30);
		s2 += "\t";
		s3 += "\t";
		dynAppend(dsText, sInformation + " " + sAlarm + "\t   " + sText + s2 + sAlias + s3 + sDpe+"\n");
		if ((sInformation != "Disable  - ") && (sAlarm != "Ok")) {
			iBad++;
			if (iBad ==1) {
				sBad = "Bad "+sDpe;
			} else {
				sBad = "Bad >1";
			}
		}
	} // end of the for each dpe

	UnSendMailServerInfo emailParams=UnSendMailServerInfo::fromCatIdx(iIndex,false);
	UnSendMailServerInfo smsParams=UnSendMailServerInfo::fromCatIdx(iIndex,true);

	if (emailParams.sendFrom=="") {
		fwException_raise(exceptionInfo, "ERROR", "_unSendMail_Report(): " + getCatStr("unSendMail", "WRONGSENDER",""), "");
		_handleExcInfo(exceptionInfo);
		return;
	}

	if (emailParams.sendTo!="") {
		_unSendMail_getText("Report", sType, sType, sLink, dsText, sTitle, sBody);
		_unSendMail_sendEmail(emailParams, sTitle, sBody);
	}

	if (smsParams.sendTo!="") {
		string status="Ok";
		if (iBad>0) status=sBad;
		sTitle=_unSendMail_getTitle("Report", status);
		_unSendMail_sendSMS(smsParams, sTitle);
	}

}

//------------------------------------------------------------------------------------------------------------------------

/** send a report with a user defined function
 *
 * @param sType  input, the mail/SMS category
 * @param sFunction  input, the user function
 */
void _unSendMail_ReportUserFunction(string sType, string sFunction)
{
	dyn_string dsDpe, dsDp_toCheck, dsReceiver, exceptionInfo, dsText;
	string sText, sTitle, sMessage, sDp, sSystem, sDpe, sAlarm, sInformation, sValue, sAck, s1, s2, sSender, sDomain, sSmtp, sLink;
	string sHost, sIPAddress, sBad, sBackup, sBody, sReceiverSMS, sReceiverMail;
	time tTime;
	int iIndex, iThread, iRes, len_check, len, i, j, len1, len10, len20, iProcess, iBad = 0;
	dyn_dyn_string dsRetFunction;


	iIndex=dynContains(g_dsType, sType);

	if (iIndex<=0) 						fwException_raise(exceptionInfo, "ERROR", "_unSendMail_ReportUserFunction(): " + getCatStr("unSendMail", "UNKNOWNCATEGORIE") + sType,"");
	if (!isFunctionDefined(sFunction)) 	fwException_raise(exceptionInfo, "ERROR", "_unSendMail_ReportUserFunction(), user function not defined: " + sFunction,"");

	if (dynlen(g_dbDisable)<iIndex  ||
		dynlen(g_dsSender)<iIndex	||
		dynlen(g_dsReceiver)<iIndex ||
		dynlen(g_dsDomain)<iIndex    )	fwException_raise(exceptionInfo, "ERROR", "_unSendMail_ReportUserFunction(): " + getCatStr("unSendMail", "PARAMETERS_MISSING"),"");

	if (_handleExcInfo(exceptionInfo)) return;

	if (g_dbDisable[iIndex])  return; // disabled - nothing to do

	dsDpe=strsplit(g_dsDpe[iIndex],"~");

	evalScript(	dsRetFunction,
					"dyn_dyn_string main(string sCat, dyn_string dsDpe) "+
					"{" +
					"return " + sFunction + "(sCat, dsDpe); "+
					"}",
				makeDynString(), sType, dsDpe);

	while(dynlen(dsRetFunction) <= UN_SEND_MAIL_BODY) dynAppend(dsRetFunction, makeDynString("", ""));

	UnSendMailServerInfo emailParams=UnSendMailServerInfo::fromCatIdx(iIndex,false);
	UnSendMailServerInfo smsParams=UnSendMailServerInfo::fromCatIdx(iIndex,true);

	if (emailParams.sendFrom=="") {
		fwException_raise(exceptionInfo, "ERROR", "_unSendMail_Report(): " + getCatStr("unSendMail", "WRONGSENDER",""), "");
		_handleExcInfo(exceptionInfo);
		return;
	}

	if (emailParams.sendTo!="") {
		sTitle = dsRetFunction[UN_SEND_MAIL_TITLE][MAIL_TITLE];
		sBody = _unSendMail_convertBody(dsRetFunction[UN_SEND_MAIL_BODY]);
		_unSendMail_sendEmail(emailParams, sTitle, sBody);
	}

	if (smsParams.sendTo!="") {
		sTitle = dsRetFunction[UN_SEND_MAIL_TITLE][SMS_TITLE];
		_unSendMail_sendSMS(smsParams, sTitle);
	}

}

//------------------------------------------------------------------------------------------------------------------------

/** Convert the body to send by email
 *
 *
 * @param dsBody  input, the content of the email
 * @return the body of the email
 */
string _unSendMail_convertBody(dyn_string dsBody)
{
	int i, len=dynlen(dsBody);
	string sReturn;

	for(i=1;i<=len;i++) sReturn = sReturn+dsBody[i]+"\n";

	return sReturn;
}

/**
 * Purpose: Worker function that sends a SMS or Email
 *
 *  Note that this function is executed in separate threads
 *   started by _unSendMail_sendEmail() an _unSendMail_sendSMS()
 *
 *	@param sReceiver: string, input, Receiver address
 *	@param sSender: string, input, Sender address
 *	@param sSmtp: string, input, smtp server name
 *	@param sDomain: string, input, domain name
 *	@param sTitle: string, input, Mail title
 *	@param sMain: string, input, Mail text
 *	@param sType: string, input, Categorie
 *
 */
void _unSendMail_Message(string sReceiver, string sSender, string sSmtp, string sDomain, string sTitle, string sMain, string sType, string sUser, string sPassword, bool bUseSsl)
{
	dyn_string dsParam, exceptionInfo, dsResult;
	string sMessage, sPath, sTime, sChar;
	file fLog;
	int iLen;

	try {

		if (sReceiver=="" || sSender=="" || sSmtp=="" || sDomain=="" || strpos(sReceiver,"@")<0 || strpos(sSender,"@")<0 ) {
			throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_WRONG_MAIL_PARAMS,sSender,sReceiver, sSmtp, sDomain));
		}

		dsParam[1]=sReceiver;
		//Replace the ';' separating the receivers by ' ' b/c since ENS-30411 sendMail is used
		strreplace( dsParam[1], UN_SENDMAIL_DELIMITER, " ");
		dsParam[2]=sSender;
		dsParam[3]=sTitle;
		dsParam[4]=sMain;

		int iRes;
		for (int i=1;i<=SEND_ATTEMPTS;i++) {
			if (i!=1) delay(DELAY_RESEND);
			//since ENS-30411 and the move from emSendMail to sendMail, sDomain is not used anymore
			sendMail(sSmtp, dsParam, iRes, sUser, sPassword, bUseSsl);
			if (iRes>=0){
				break;
			}
		}

		//save Result
		sTime=formatTime("%Y-%m-%d %H:%M:%S", getCurrentTime());
		if (iRes>=0) {
			fwException_raise(exceptionInfo, "EXPERTINFO", sTitle + " successfully sent to " + sReceiver,"");
			_handleExcInfo(exceptionInfo);
		} else {
			throw(makeError("unSendMail",PRIO_SEVERE,ERR_CONTROL,EXC_SEND_FAIL,sTitle,sReceiver,iRes));
		}


	} catch {

		_handleErrClass(getLastException());

	} finally {

		//Write Log
		strrtrim(sReceiver,";");
		sPath = getPath(LOG_REL_PATH) + sType + "_to_" + sReceiver + ".txt";
		fLog = fopen(sPath, "w");
		if (fLog != 0) {
			fputs(sTime+"/n", fLog);
			fputs(sMain, fLog);
			fclose(fLog);
		}
	}

}

/**
 * Purpose: send a Email when system starts
 *
 *	@param dsSettings: dyn_string, input, Settings
 *	@param dbDisable: dyn_bool, input, Categorie Disable?
 *
 */
void unSendMail_SendInitNotifications()
{
	string sType="Starting";

	for (int i=1;i<=dynlen(g_dsType);i++) {				//for each Category

		string category=g_dsType[i];
		// the key used in mappings
		string cat=strtolower(category);

		g_diDayCounter[cat]=0;
		g_diToday[cat]=0;

		if (g_dbDisable[i])  continue;

		// check if we need to send anything at startup
		string sStartup = "";
		bool bStartup = false;
		dyn_string exceptionInfo;
		unProcess_getReportCategoryFunction(getSystemName(), g_dsType[i]+UN_SEND_MAIL_KEY_CATEGORY_STARTUP, sStartup, exceptionInfo);
		if (dynlen(exceptionInfo) <=0) bStartup = (bool)sStartup;
		if (!bStartup) continue;

		string sLink =g_dsLink[i];

		UnSendMailServerInfo emailParams = UnSendMailServerInfo::fromCatIdx(i,false);
		UnSendMailServerInfo smsParams   = UnSendMailServerInfo::fromCatIdx(i,true);

		dyn_string dsText; // empty text
		string sTitle, sMain;
		_unSendMail_getText(sType, "Information", category, sLink, dsText, sTitle, sMain);

		if (emailParams.sendTo!="") _unSendMail_sendEmail(emailParams, sTitle, sMain, sType);
		if (smsParams.sendTo!="")   _unSendMail_sendSMS(smsParams, sTitle, sType);
	}
}

/**
 * Purpose: get only Title text
 *
 *	@param sMessageType: string, input, Message type
 *	@param sInfo: string, input, info
 *	@param sTimestamp: the timestamp; if left empty (default), current timestap will be used
 *
 */
string _unSendMail_getTitle(string sMessageType,
							string sInfo,
							string sTimestamp="")
{
	if (sTimestamp=="") {
		time t=getCurrentTime();
		sTimestamp=formatTime("%Y-%m-%d %H:%M:%S", t,".%03d");
	}

	string sTitle = sMessageType + "/" + sInfo + ">" + sTimestamp;
	return sTitle;
}

/** Gets the custom message for specified category

	@param sCategory : the name of the category
	@return (string) the custom message configured for the category
 */
string _unSendMail_getCustomMessage(const string& sCategory)
{
	dyn_string dsCustomText, dsTemp;
	string sCustomText;

	dpGet("_unSendMail.config.mail_text",dsCustomText);
	for (int i = 1; i<= dynlen(dsCustomText);i++) {
		dsTemp = strsplit(dsCustomText[i],"~");
		if (dynlen(dsTemp) != 2) continue;
		if (dsTemp[1] == sCategory) {
			sCustomText = dsTemp[2];
			break;
		}
	}
	return sCustomText;
}

/**
 * Purpose: get Title and Main text
 *
 *	@param sMessageType: string, input, Message type
 *	@param sInfo: string, input, info
 *	@param sCategory: string, input, Category name
 *	@param sLink: string, input, HTML Link
 *	@param dsBody: dyn_string, input, Body Text
 *	@param sTitle: string, output, Title Text
 *	@param sMain: string, output, Main Text
 *	@param sTime: the timestamp; if left empty (default), current timestap will be used
 *
 */
void _unSendMail_getText(string sMessageType,
						string sInfo,
						string sCategory,
						string sLink,
						dyn_string dsBody,
						string &sTitle,
						string &sMain,
						string sTimestamp="")
{
	dyn_string dsMain;

	//init
	sTitle="";
	sMain="";
	dynClear(dsMain);

	string sHost, sIPAddress;
	string sSystem=getSystemName();
	unGenericDpFunctions_getHostName(sSystem, sHost, sIPAddress);

	if (sTimestamp=="") {
		time t=getCurrentTime();
		sTimestamp=formatTime("%Y-%m-%d %H:%M:%S", t,".%03d");
	}

	// re-format the link; convert on-the-flight from the old LEMON to new COSMOS URL
  sLink=unSendMail_getHTMLLink();

	//format the title
	sTitle=_unSendMail_getTitle(sMessageType, sInfo, sTimestamp);

	// get custom message
	dyn_string sCustomText = _unSendMail_getCustomMessage(sCategory);

	//format the main text
	//dynAppend(dsMain, "Type: " + sMessageType + "\n");
	dynAppend(dsMain, dsBody);
	dynAppend(dsMain, "\nSystem: " +  sSystem + "\n");
	dynAppend(dsMain, "Category: " + sCategory + "\n");
	dynAppend(dsMain, "DS Host Name: " +  sHost + "\n");
	dynAppend(dsMain, "DS IP Address: " + sIPAddress  + "\n\n");
	dynAppend(dsMain, sLink	+"\n\n");
	dynAppend(dsMain, sCustomText + "\n\t\t------------------------\n\n");
	dynAppend(dsMain, "\n");

	for (int i=1; i<=dynlen(dsMain); i++) sMain=sMain+dsMain[i];
}

/**
 * Purpose: check if the key is defined in the Globals mapping
 *
 *	@param sCategory: string, input, Key name
 *
 */
void _unSendMail_checkMapping(const string sCategory)
{
	string sKey=strtolower(sCategory);

	if (!mappingHasKey(g_diDayCounter, sKey)) g_diDayCounter[sKey]=0;
	if (!mappingHasKey(g_diToday,      sKey)) g_diToday[sKey]=0;
}

/**
 * Purpose: convert SystemAlarm's value to corresponding text description (LHCLogging/Laser/CMWServer)
 *
 *	@param sDpe: string, input, Dpe name
 *	@param sValue: string, input, value
 *
 */
string _unSendMail_convertValueToText(const string sDpe, const anytype sValue)
{
	string sText;

	if (sValue=="")  return "(no value)";

	string cat="";
	if      (strpos(sDpe,"LHCLogging")>0 && getPath(MSG_REL_PATH,"unLHCLogging.cat") !="" ) cat="unLHCLogging";
	else if (strpos(sDpe,"Laser"     )>0 && getPath(MSG_REL_PATH,"unLaser.cat")      !="" ) cat="unLaser";
	else if (strpos(sDpe,"CMWServer" )>0 && getPath(MSG_REL_PATH,"unCMWServer.cat") !="" )  cat="unCMWServer";

	if (cat!="") {
		return getCatStr(cat, sValue) + " (" + sValue + ")";
	}

	return dpValToString(sDpe, sValue);
}
//@}
